"""Test :func:`sievemgr.backup`."""

#
# Copyright 2024  Odin Kroeger
#
# This file is part of SieveManager.
#
# SieveManager is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# SieveManager is distributed in the hope that it will be useful,
# but WITHOUT ALL WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SieveManager. If not, see <https://www.gnu.org/licenses/>.
#

# pylint: disable=missing-class-docstring,missing-function-docstring


#
# Modules
#

import os
import pathlib
import shutil
import sys
import tempfile
import unittest

from os import path

sys.path.append(path.realpath(pathlib.Path(__file__).parents[1]))

from sievemgr import backup, readdir


#
# Cases
#

MOCK_ERRORS: tuple[tuple[str, int, type[Exception]], ...] = (
    ('foo', -1, ValueError),
    ('bar', 1, KeyError),
    ('bar', 2, KeyError)
)


FILE_ERRORS: tuple[tuple[str, int, type[Exception]], ...] = (
    ('foo', -1, ValueError),
    ('bar', 1, FileNotFoundError),
    ('bar', 2, FileNotFoundError)
)


#
# Tests
#

class TestMockup(unittest.TestCase):
    def backup(self, fname: str, keep: int):
        def getfiles() -> tuple[str, ...]:
            return tuple(self.filesystem)
        def copy(src: str, targ: str):
            if src not in self.filesystem:
                raise KeyError(src)
            self.filesystem.add(targ)
        def remove(fname: str):
            self.filesystem.remove(fname)
        return backup(fname, keep, getfiles, copy, remove)

    def test_error(self):
        for i, (fname, keep, exctype) in enumerate(MOCK_ERRORS):
            with self.subTest(i=i, keep=keep, exctype=exctype):
                self.filesystem = {'foo'}
                try:
                    self.backup(fname, keep)
                # pylint: disable=broad-exception-caught
                except Exception as err:
                    self.assertIsInstance(err, exctype)
                    continue
                self.fail(f'expected {exctype}')

    def test_noop(self):
        self.filesystem = {'foo'}
        self.backup('foo', 0)
        self.assertEqual(self.filesystem, {'foo'})

    def test_simple(self):
        for i in range(1, 12):
            with self.subTest(i=i):
                self.filesystem = {'foo'}
                for _ in range(i):
                    self.backup('foo', 1)
                self.assertEqual(len(self.filesystem), 2)
                self.assertIn('foo', self.filesystem)
                self.assertIn('foo~', self.filesystem)

    def test_numbered(self):
        for i in range(2, 11):
            with self.subTest(i=i):
                self.filesystem = {'foo'}
                delta = 2
                count = i + delta
                for _ in range(count):
                    self.backup('foo', i)
                self.assertEqual(len(self.filesystem), i + 1)
                self.assertIn('foo', self.filesystem)
                for j in range(delta, count):
                    self.assertIn(f'foo.~{j + 1}~', self.filesystem)

    filesystem: set[str] = set()


class TestFile(unittest.TestCase):
    def backup(self, fname: str, keep: int):
        return backup(fname, keep, lambda: readdir(''), shutil.copy, os.remove)

    def test_error(self):
        oldwd = os.getcwd()
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                os.chdir(tmpdir)
                with open('foo', 'w', encoding='utf8'):
                    pass
                for i, (fname, keep, exctype) in enumerate(FILE_ERRORS):
                    with self.subTest(i=i, fname=fname,
                                      keep=keep, exctype=exctype):
                        try:
                            self.backup(fname, keep)
                        # pylint: disable=broad-exception-caught
                        except Exception as err:
                            self.assertIsInstance(err, exctype)
                            continue
                        self.fail(f'expected {exctype}')

            finally:
                os.chdir(oldwd)

    def test_noop(self):
        oldwd = os.getcwd()
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                os.chdir(tmpdir)
                with open('foo', 'w', encoding='utf8'):
                    pass
                self.backup('foo', 0)
                self.assertEqual(len(list(readdir('.', path.isfile))), 1)
            finally:
                os.chdir(oldwd)

    def test_simple(self):
        oldwd = os.getcwd()
        for i in range(1, 12):
            with self.subTest(i=i):
                with tempfile.TemporaryDirectory() as tmpdir:
                    try:
                        os.chdir(tmpdir)
                        with open('foo', 'w', encoding='utf8'):
                            pass
                        for _ in range(i):
                            self.backup('foo', 1)
                        files = list(readdir('', path.isfile))
                        self.assertEqual(len(files), 2)
                        self.assertIn('foo', files)
                        self.assertIn('foo~', files)
                    finally:
                        os.chdir(oldwd)

    def test_numbered(self):
        oldwd = os.getcwd()
        for i in range(2, 12):
            with self.subTest(i=i):
                delta = 2
                count = i + delta
                with tempfile.TemporaryDirectory() as tmpdir:
                    try:
                        os.chdir(tmpdir)
                        with open('foo', 'w', encoding='utf8'):
                            pass
                        for _ in range(count):
                            self.backup('foo', i)
                        files = list(readdir('', path.isfile))
                        self.assertEqual(len(files), i + 1)
                        self.assertIn('foo', files)
                        for j in range(delta, count):
                            self.assertIn(f'foo.~{j + 1}~', files)
                    finally:
                        os.chdir(oldwd)


#
# Boilerplate
#

if __name__ == '__main__':
    unittest.main()



"""Test :class:`sievemgr.BaseShell`."""

#
# Copyright 2024  Odin Kroeger
#
# This file is part of SieveManager.
#
# SieveManager is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# SieveManager is distributed in the hope that it will be useful,
# but WITHOUT ALL WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SieveManager. If not, see <https://www.gnu.org/licenses/>.
#

# pylint: disable=missing-class-docstring,missing-function-docstring


#
# Modules
#

import io
import itertools
import os
import pathlib
import shutil
import sys
import unittest

from typing import Final, Iterator, Optional, Sequence
from unittest.mock import patch

sys.path.append(os.path.realpath(pathlib.Path(__file__).parents[1]))

import sievemgr

from sievemgr import BaseShell, ConfirmEnum, ShellPattern, ShellWord
from . import Test, runtests


#
# Silence the bell
#

def silence():
    pass

sievemgr.bell = silence


#
# Functions
#

def addspaces(s: str, n: int = 3) -> Iterator[str]:
    for i in range(n):
        yield s + ' ' * i
        yield ' ' * i + s


def mixcase(s: str) -> Iterator[str]:
    yield from map(''.join, itertools.product(*zip(s.upper(), s.lower())))


#
# Globals
#

COMPLETIONS: Final[tuple[Test[Optional[str]], ...]] = (
    # Empty
    *[(('', '', i), cmd)
      for i, cmd in enumerate(('bar ', 'exit ', 'foo ', 'help ', None))],

    # Commands
    *[(('', cmd[0:i], j), cmd if j == 0 else None)
      for cmd in ('bar ', 'exit ', 'foo ', 'help ')
      for i in range(1, len(cmd))
      for j in (0, 1)],

    # Arguments
    (('exit ', '', 0), None),
    (('exit ', '', 1), None),
    (('help ', '', 0), 'bar '),
    (('help ', '', 1), 'exit '),
    (('help ', '', 2), 'foo '),
    (('help ', '', 3), 'help '),
    (('help ', '', 4), None),
    (('foo ', '', 0), 'bar '),
    (('foo ', '', 1), 'baz '),
    (('foo ', '', 2), 'foo '),
    (('foo ', '', 3), None),
    (('foo ', '', 4), None),
    (('foo ', 'b', 0), 'bar '),
    (('foo ', 'b', 1), 'baz '),
    (('foo ', 'b', 2), None),
    (('foo ', 'b', 3), None),
    (('foo ', 'ba', 0), 'bar '),
    (('foo ', 'ba', 1), 'baz '),
    (('foo ', 'ba', 2), None),
    (('foo ', 'ba', 3), None),
    (('foo ', 'bar', 0), 'bar '),
    (('foo ', 'bar', 1), None),
    (('foo ', 'bar', 2), None),
    (('foo ', 'baz', 0), 'baz '),
    (('foo ', 'baz', 1), None),
    (('foo ', 'baz', 2), None),
    (('foo ', 'bar ', 0), None),
    (('foo ', 'bar ', 1), None),
    (('foo ', 'baz ', 0), None),
    (('foo ', 'baz ', 1), None),
    (('foo ', 'f', 0), 'foo '),
    (('foo ', 'f', 1), None),
    (('foo ', 'f', 2), None),
    (('foo ', 'fo', 0), 'foo '),
    (('foo ', 'fo', 1), None),
    (('foo ', 'fo', 2), None),
    (('foo ', 'foo', 0), 'foo '),
    (('foo ', 'foo', 1), None),
    (('foo ', 'foo', 2), None),
    (('foo ', 'foo ', 0), None),
    (('foo ', 'foo ', 1), None),
    (('foo foo ', '', 0), 'quuuux'),
    (('foo foo ', '', 1), 'quuux'),
    (('foo foo ', '', 2), 'quux'),
    (('foo foo ', '', 3), None),
    (('foo foo ', 'q', 0), 'quuuux'),
    (('foo foo ', 'q', 1), 'quuux'),
    (('foo foo ', 'q', 2), 'quux'),
    (('foo foo ', 'q', 3), None),
    (('foo foo ', 'qu', 0), 'quuuux'),
    (('foo foo ', 'qu', 1), 'quuux'),
    (('foo foo ', 'qu', 2), 'quux'),
    (('foo foo ', 'qu', 3), None),
    (('foo foo ', 'quu', 0), 'quuuux'),
    (('foo foo ', 'quu', 1), 'quuux'),
    (('foo foo ', 'quu', 2), 'quux'),
    (('foo foo ', 'quu', 3), None),
    (('foo foo ', 'quuu', 0), 'quuuux'),
    (('foo foo ', 'quuu', 1), 'quuux'),
    (('foo foo ', 'quuu', 2), None),
    (('foo foo ', 'quuuu', 0), 'quuuux'),
    (('foo foo ', 'quuuu', 1), None),
    (('foo foo ', 'quuuuu', 0), None),
    (('foo foo ', 'quuuux', 0), 'quuuux'),
    (('foo foo ', 'quuuux', 1), None),
    (('foo foo ', 'quuuux ', 0), None),

    # Patterns
    (('foo ', '*', 0), 'bar '),
    (('foo ', '*', 1), 'baz '),
    (('foo ', '*', 2), 'foo '),
    (('foo ', '*', 3), None),
    (('foo ', 'f*', 0), 'foo '),
    (('foo ', 'f*', 1), None),
    (('foo ', 'fo*', 0), 'foo '),
    (('foo ', 'fo*', 1), None),
    (('foo ', 'f*o', 0), 'foo '),
    (('foo ', 'f*o', 1), None),
    (('foo ', 'fo?', 0), 'foo '),
    (('foo ', 'fo?', 1), None),
    (('foo ', 'f?o', 0), 'foo '),
    (('foo ', 'f?o', 1), None),
    (('foo ', 'fo[mno]', 0), 'foo '),
    (('foo ', 'fo[mno]', 1), None),
    (('foo ', 'f[o]o', 0), 'foo '),
    (('foo ', 'f[o]o', 1), None),
    (('foo ', 'f[np]o', 0), None)
)


CONFIRMATIONS: Final[tuple[Test[ConfirmEnum], ...]] = (
    # Defaults
    *[((' ' * i, enum), enum)
      for enum in ConfirmEnum
      for i in range(3)],

    # Yes, no, all, and none
    *[((z, ConfirmEnum.YES), ConfirmEnum.NO)
      for x in ('n', 'no')
      for y in mixcase(x)
      for z in addspaces(y)],
    *[((z, ConfirmEnum.NO), ConfirmEnum.YES)
      for x in ('y', 'yes')
      for y in mixcase(x)
      for z in addspaces(y)],
    *[((y, ConfirmEnum.YES, True), ConfirmEnum.NONE)
      for x in mixcase('none')
      for y in addspaces(x)],
    *[((y, ConfirmEnum.NO, True), ConfirmEnum.ALL)
      for x in mixcase('all')
      for y in addspaces(x)],

    # Errors
    *[((z, ConfirmEnum.NO), ValueError)
      for x in ('all', 'none')
      for y in mixcase(x)
      for z in addspaces(y)],
    *[((z, ConfirmEnum.NO, False), ValueError)
      for x in ('all', 'none')
      for y in mixcase(x)
      for z in addspaces(y)],
    *[((z, ConfirmEnum.NO), ValueError)
      for x in ('foo', 'bar', 'baz')
      for y in mixcase(x)
      for z in addspaces(y)],
    *[((z, ConfirmEnum.NO, False), ValueError)
      for x in ('foo', 'bar', 'baz')
      for y in mixcase(x)
      for z in addspaces(y)],
)


LINES: Final[tuple[Test[list[ShellWord]], ...]] = (
    # Simple
    (('foo',), ['foo']),
    (('foo bar',), ['foo', 'bar']),
    (('foo bar baz',), ['foo', 'bar', 'baz']),

    # Empty
    (('',), []),
    ((' ',), []),
    (('        ',), []),
    (('\t',), []),
    (('\t\n     ',), []),
    (('#',), []),
    (('# foo',), []),
    (('    # foo',), []),
    (('\t\n    # foo\\\nfoo',), []),

    # Leading and trailing whitespace
    ((' foo',), ['foo']),
    ((' foo bar',), ['foo', 'bar']),
    ((' foo bar baz',), ['foo', 'bar', 'baz']),
    (('foo ',), ['foo']),
    (('foo bar ',), ['foo', 'bar']),
    (('foo bar baz ',), ['foo', 'bar', 'baz']),
    ((r'foo\ ',), ['foo ']),
    ((r'foo bar\ ',), ['foo', 'bar ']),
    ((r'foo bar baz\ ',), ['foo', 'bar', 'baz ']),

    # Quotes
    (('"foo bar"',), ['foo bar']),
    (("'foo bar'",), ['foo bar']),

    # Escaped sequences
    ((r'foo\ bar',), ['foo bar']),
    ((r'\"foo bar"',), ['"foo', 'bar']),
    ((r"\'foo bar'",), ["'foo", 'bar']),

    # Escaped and quoted patterns
    ((r'\*',), ['*']),
    ((r'\?',), ['?']),
    ((r'\[',), ['[']),
    ((r'"*"',), ['*']),
    ((r'"?"',), ['?']),
    ((r'"["',), ['[']),
    ((r"'*'",), ['*']),
    ((r"'?'",), ['?']),
    ((r"'['",), ['[']),
    ((r'foo \*',), ['foo', '*']),
    ((r'\?\bar',), ['?bar']),
    ((r'\[foo]',), ['[foo]']),
    ((r'"foo *"',), ['foo *']),
    ((r'"b?r"',), ['b?r']),
    ((r'"[bar" ]',), ['[bar', ']']),
    ((r"'fo*' fo",), ['fo*', 'fo']),
    ((r"bar '?'",), ['bar', '?']),
    ((r"'['",), ['[']),
)


PATTERNSRAW: Final[tuple[Test[list[ShellWord]], ...]] = (
    (('* bar',), [ShellPattern('*'), 'bar']),
    (('? bar',), [ShellPattern('?'), 'bar']),
    (('[foo] bar',), [ShellPattern('[foo]'), 'bar']),
    ((r'\* bar',), [r'*', 'bar']),
    ((r'\? bar',), [r'?', 'bar']),
    ((r'\[foo] bar',), [r'[foo]', 'bar']),
    ((r'\** bar',), [ShellPattern(r'[*]*'), 'bar']),
    ((r'\?? bar',), [ShellPattern(r'[?]?'), 'bar']),
    ((r'\[foo][bar] baz',), [ShellPattern(r'[[]foo][bar]'), 'baz'])
)


PATTERNSEXP: Final[tuple[Test[list[ShellWord]], ...]] = (
    (('foo * ',), ['foo', 'bar', 'baz', 'foo']),
    (('foo ?oo',), ['foo', 'foo']),
    (('foo [f]oo',), ['foo', 'foo']),
    ((r'foo \*',), ['foo', '*']),
    ((r'foo \?',), ['foo', '?']),
    ((r'foo \[foo]',), ['foo', '[foo]']),
    ((r'foo \**',), ValueError),
    ((r'foo \??',), ValueError),
    ((r'foo \[foo][foo]',), ValueError)
)


USAGE: Final[tuple[Test[Optional[str]], ...]] = (
    ((None,), None),
    (('',), None),
    (('-',), None),
    ((' -',), None),
    (('- ',), None),
    ((' - ',), None),
    (('- foo',), None),
    (('-- foo',), 'usage: -'),
    (('frob',), 'usage: frob'),
    (('frob foo - frobnicate foo',), 'usage: frob foo'),
    (('frob [-o] foo - frobnicate foo',), 'usage: frob [-o] foo'),
)


WORDS: Final[tuple[Test[str], ...]] = (
    (((), 0), ''),
    ((('',), 0), '\n'),
    (((' ',), 0), '\n'),
    (((), 1), ''),
    ((('',), 1), '\n'),
    (((' ',), 1), '\n'),
    (((), 2), ''),
    ((('',), 2), '\n'),
    (((' ',), 2), '\n'),
    ((('foo',), 3), 'foo\n'),
    ((('foo', 'bar'), 3), 'foo\nbar\n'),
    ((('foo', 'bar'), 4), 'foo\nbar\n'),
    ((('foo', 'bar'), 6), 'foo\nbar\n'),
    ((('foo', 'bar'), 7), 'foo bar\n')
)


#
# Classes
#

class MockReadline():
    def __init__(self, line: str):
        for func in self.functions:
            self.mockups[func] = patch(func, autospec=True)
        self.line = line

    def __enter__(self) -> 'MockReadline':
        ctx = {func.removeprefix('readline.'): mockup.__enter__()
               for func, mockup in self.mockups.items()}
        ctx['get_begidx'].return_value = None
        ctx['get_line_buffer'].return_value = self.line
        return self

    def __exit__(self, exctype, excvalue, traceback):
        for mockup in self.mockups.values():
            mockup.__exit__(exctype, excvalue, traceback)

    functions = ('readline.get_begidx', 'readline.get_line_buffer')
    mockups: dict = {}


class TestShellMixin():
    def do_foo(self):
        pass

    def do_bar(self):
        pass

    def do_(self):
        pass

    def complete_foo(self, argidx, _) -> tuple[tuple[str, bool], ...]:
        if argidx == 0:
            return (('impossible', False),)
        if argidx == 1:
            return tuple((s, True) for s in ('foo', 'bar', 'baz'))
        if argidx == 2:
            return tuple((s, False) for s in ('quux', 'quuux', 'quuuux'))
        return ()


#
# Tests
#

class TestBaseShell(unittest.TestCase):
    # Wrappers
    @classmethod
    def columnise(cls, words: Sequence[str],
                  width: int = shutil.get_terminal_size().columns) -> str:
        buf = io.StringIO()
        cls.basecls.columnise(words, file=buf, width=width)
        return buf.getvalue()

    def complete(self, line: str, text: str, n: int) -> Optional[str]:
        with MockReadline(line=line):
            return self.shell.complete(text, n)

    @classmethod
    def confirm(cls, line: str, default: ConfirmEnum = ConfirmEnum.NO,
                multi: bool = False) -> ConfirmEnum:
        confirm = cls.basecls.confirm
        with patch('sievemgr.TermIO.readline', autospec=True) as readline:
            with patch('sievemgr.TermIO.write', autospec=True):
                readline.return_value = line
                return confirm('', default=default, multi=multi, attempts=1)

    def expand(self, line: str) -> list[ShellWord]:
        return self.shell.expand(line)

    @classmethod
    def getargs(cls, line: str) -> list[ShellWord]:
        with MockReadline(line=line):
            return cls.basecls.getargs()

    @classmethod
    def getusage(cls, docstring: Optional[str]) -> Optional[str]:
        def func():
            pass
        func.__doc__ = docstring
        return cls.basecls.getusage(func)

    @classmethod
    def split(cls, line: str) -> list[ShellWord]:
        return cls.basecls.split(line)

    # Setup
    def setUp(self):
        self.shellcls = type('TestShell', (self.basecls, TestShellMixin), {})
        self.shell = self.shellcls()

    # Tests
    def test_init(self):
        base = self.basecls()
        assert {'exit', 'help'} <= set(base.commands)
        sh = self.shellcls()
        assert {'bar', 'exit', 'foo', 'help'} <= set(sh.commands)

    def test_columnise(self):
        return runtests(self, self.columnise, WORDS)

    def test_complete(self):
        return runtests(self, self.complete, COMPLETIONS)

    @unittest.skipIf(os.getenv('QUICKTEST'), 'slow')
    def test_confirm(self):
        try:
            self.basecls.confirm('foo', attempts=0)
        except AssertionError:
            return
        self.fail('expected AssertionError')
        return runtests(self, self.confirm, CONFIRMATIONS)

    # TODO
    # def test_enter

    # TODO
    # def test_execute

    # TODO
    # def test_executeline

    # TODO
    # def test_executescript

    def test_expand(self):
        return runtests(self, self.expand, LINES + PATTERNSEXP)

    @unittest.skipIf(os.getenv('QUICKTEST'), 'slow')
    def test_getargs(self):
        return runtests(self, self.getargs, LINES + PATTERNSRAW)

    def test_getcommands(self):
        self.assertEqual(list(self.shell.getcommands()),
                         ['bar', 'exit', 'foo', 'help'])

    def test_getprompt(self):
        self.assertEqual(self.shell.getprompt(), '> ')

    def test_getusage(self):
        return runtests(self, self.getusage, USAGE)

    def test_split(self):
        return runtests(self, self.split, LINES + PATTERNSRAW)

    def test_do_exit(self):
        try:
            self.shell.do_exit()
        except StopIteration:
            return
        self.fail('expected StopIteration')

    # TODO
    # def test_do_help

    basecls: type[BaseShell] = BaseShell

    shellcls: type[BaseShell]


# TODO:
# * Test if aliases are resolved
# * Test if help messages are generated correctly
# * Add tests for ShellPattern


#
# Boilerplate
#

if __name__ == '__main__':
    unittest.main()



"""Test :class:`sievemgr.LogIOWrapper`."""

#
# Copyright 2024  Odin Kroeger
#
# This file is part of SieveManager.
#
# SieveManager is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# SieveManager is distributed in the hope that it will be useful,
# but WITHOUT ALL WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SieveManager. If not, see <https://www.gnu.org/licenses/>.
#

# pylint: disable=missing-class-docstring,missing-function-docstring


#
# Modules
#

import logging
import os
import pathlib
import sys
import tempfile
import unittest

sys.path.append(os.path.realpath(pathlib.Path(__file__).parents[1]))

from sievemgr import LogIOWrapper


#
# Tests
#

class TestLogIOWrapper(unittest.TestCase):
    """Tests for :class:`LogIOWrapper`."""

    def setUp(self):
        progname = os.path.basename(sys.argv[0])
        logging.basicConfig(format=f'{progname}: %(message)s')
        self.file = LogIOWrapper(tempfile.TemporaryFile())

    def tearDown(self):
        self.file.close()

    def test_close(self):
        file = LogIOWrapper(tempfile.TemporaryFile())
        file.close()
        self.assertTrue(file.closed)

    def test_closed(self):
        self.assertFalse(self.file.closed)

    def test_fileno(self):
        self.assertTrue(self.file.fileno() > 0)

    def test_flush(self):
        self.file.flush()

    def test_isatty(self):
        self.assertFalse(self.file.isatty())
        with open('/dev/tty', mode='rb') as tty:
            file = LogIOWrapper(tty)
            self.assertTrue(file.isatty())

    def test_iter(self):
        file = self.file
        pos = file.tell()
        write = (b'foo\n', b'bar\n', b'baz\n')
        file.writelines(write)
        file.seek(pos)
        read = tuple(line for line in file)
        self.assertEqual(read, write)

    def test_read(self):
        pos = self.file.tell()
        self.file.write(b'foo\n')
        self.file.seek(pos)
        self.assertEqual(self.file.read(), b'foo\n')

    def test_readinto(self):
        pos = self.file.tell()
        write = b'foo\n'
        self.file.write(write)
        self.file.seek(pos)
        read = bytearray(len(write))
        self.file.readinto(read)
        self.assertEqual(read, write)

    def test_readable(self):
        self.assertTrue(self.file.readable())

    def test_readline(self):
        pos = self.file.tell()
        self.file.write(b'foo\n')
        self.file.seek(pos)
        self.assertEqual(self.file.readline(), b'foo\n')

    def test_readlines(self):
        pos = self.file.tell()
        self.file.write(b'foo\nbar\n')
        self.file.seek(pos)
        self.assertEqual(self.file.readlines(), [b'foo\n', b'bar\n'])

    def test_seek(self):
        pos = self.file.tell()
        self.file.seek(0)
        self.assertEqual(self.file.tell(), 0)
        self.file.seek(pos)

    def test_seekable(self):
        self.assertTrue(self.file.seekable())

    def test_truncate(self):
        self.file.write(b'foo')
        self.file.seek(0)
        self.assertEqual(self.file.truncate(), 0)
        self.file.seek(0, os.SEEK_END)
        self.assertEqual(self.file.tell(), 0)

    def test_writable(self):
        return self.assertTrue(self.file.writable)

    def test_writelines(self):
        lines = [b'foo\n', b'bar\n']
        pos = self.file.tell()
        self.file.writelines(lines)
        self.file.seek(pos)
        self.assertEqual(self.file.readlines(), lines)


#
# Boilerplate
#

logging.getLogger(__name__).addHandler(logging.NullHandler())

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Tests for :class:`sievemgr.BaseAuth`."""

#
# Copyright 2024  Odin Kroeger
#
# This file is part of SieveManager.
#
# SieveManager is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# SieveManager is distributed in the hope that it will be useful,
# but WITHOUT ALL WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SieveManager. If not, see <https://www.gnu.org/licenses/>.
#

# pylint: disable=missing-class-docstring, missing-function-docstring


#
# Modules
#

import os
import pathlib
import sys
import unittest

from typing import Final, Iterator, TypeVar, Union
from unicodedata import normalize

sys.path.append(os.path.realpath(pathlib.Path(__file__).parents[1]))

from sievemgr import BaseAuth
from . import Test, makerunner


#
# Types
#

T = TypeVar('T')


#
# Helpers
#

def mapchars(first: int, last: int, obj: Union[T, type[Exception]]) \
        -> Iterator[Test[T]]:
    for i in range(first, last + 1):
        yield ((chr(i),), obj)


#
# Globals
#

USERNAMES: set[str] = {
    normalize(form, user)   # type: ignore[arg-type]
    for user in (
        '0123', 'user', 'UserName', 'user_name', 'tim',
        'user@localhost', 'user@foo.example', 'user@xn--bcher-kva.example',
        'فلان', 'فلانة', 'فلان الفلاني', 'فلانة الفلانية', 'علان',
        'ܐܸܢܵܐ', 'ܦܠܵܢ', 'ܦܠܵܢܝܼܬ݂ܵܐ',
        'ইয়ে', 'অমুক', 'ফলনা', 'ফলানা',
        'Иван', 'Драган', 'Петкан',
        '陳大文', '陳小明', '路人甲', 'Chris Wong',
        'mikälie',
        'Γιάννης Παπαδόπουλος',
        'מה-שמו', 'משה', 'יוֹסִי', 'כהן', 'ישראל ישראלי', 'ישראלה ישראלי',
        'פלוני', 'אלמוני', 'פלוני אלמוני', 'ראובן', 'שמעון',
        'Jón Jónsson', 'Jóna Jónsdóttir',
        '홍길동', '아무개', '철수', '영희', '某人',
        '李某', '张三', '張三', '李四', '王五',
        'فلانی', 'طرف', 'يارو',
        'такой-то', 'имярек', 'Пётр', 'Сидор',
        'Иван Петрович Сидоров', 'Вася Пупкин', 'Василий Пупкин',
        'Petar Petrović', 'Pera Perić',
        'askurđel',
        'Sarı Çizmeli Mehmet Ağa',
        'невідомий',
        'Nguyễn Văn A', 'Nguyễn Thị A'
    )
    for form in ('NFC', 'NFKC', 'NFD', 'NFKD')
}


STRINGS: Final[tuple[Test[str], ...]] = (
    # Simple tests
    (('',), ''),
    (('foo',), 'foo'),
    (('\n',), ValueError),
    (('\r',), ValueError),
    (('\0',), ValueError),

    # Edge cases
    ((' ',), ' '),
    (('  ',), '  '),
    (('0',), '0'),
    (('00',), '00'),

    # Usernames
    *[((u,), normalize('NFC', u)) for u in USERNAMES],

    # Unassigned code points
    ((chr(0x0221),), ValueError),
    *mapchars(0x0234, 0x024f, ValueError),
    *mapchars(0x02ae, 0x02af, ValueError),
    *mapchars(0x02ef, 0x02ff, ValueError),
    *mapchars(0x0350, 0x035f, ValueError),
    *mapchars(0x0370, 0x0373, ValueError),
    *mapchars(0x0376, 0x0379, ValueError),
    *mapchars(0x037b, 0x037d, ValueError),
    *mapchars(0x037f, 0x0383, ValueError),
    ((chr(0x038b),), ValueError),
    ((chr(0x038d),), ValueError),
    ((chr(0x03a2),), ValueError),
    ((chr(0x03cf),), ValueError),
    *mapchars(0x03f7, 0x03ff, ValueError),
    ((chr(0x0487),), ValueError),
    ((chr(0x04cf),), ValueError),
    *mapchars(0x04f6, 0x04f7, ValueError),
    *mapchars(0x04fa, 0x04ff, ValueError),
    *mapchars(0x0510, 0x0530, ValueError),
    *mapchars(0x0557, 0x0558, ValueError),
    ((chr(0x0560),), ValueError),
    ((chr(0x0588),), ValueError),
    *mapchars(0x058b, 0x0590, ValueError),
    ((chr(0x05a2),), ValueError),
    ((chr(0x05ba),), ValueError),
    *mapchars(0x05c5, 0x05cf, ValueError),
    *mapchars(0x05eb, 0x05ef, ValueError),
    *mapchars(0x05f5, 0x060b, ValueError),
    *mapchars(0x060d, 0x061a, ValueError),
    *mapchars(0x061c, 0x061e, ValueError),
    ((chr(0x0620),), ValueError),
    *mapchars(0x063b, 0x063f, ValueError),
    *mapchars(0x0656, 0x065f, ValueError),
    *mapchars(0x06ee, 0x06ef, ValueError),
    ((chr(0x06ff),), ValueError),
    ((chr(0x070e),), ValueError),
    *mapchars(0x072d, 0x072f, ValueError),
    *mapchars(0x074b, 0x077f, ValueError),
    *mapchars(0x07b2, 0x0900, ValueError),
    ((chr(0x0904),), ValueError),
    *mapchars(0x093a, 0x093b, ValueError),
    *mapchars(0x094e, 0x094f, ValueError),
    *mapchars(0x0955, 0x0957, ValueError),
    *mapchars(0x0971, 0x0980, ValueError),
    ((chr(0x0984),), ValueError),
    *mapchars(0x098d, 0x098e, ValueError),
    *mapchars(0x0991, 0x0992, ValueError),
    ((chr(0x09a9),), ValueError),
    ((chr(0x09b1),), ValueError),
    *mapchars(0x09b3, 0x09b5, ValueError),
    *mapchars(0x09ba, 0x09bb, ValueError),
    ((chr(0x09bd),), ValueError),
    *mapchars(0x09c5, 0x09c6, ValueError),
    *mapchars(0x09c9, 0x09ca, ValueError),
    *mapchars(0x09ce, 0x09d6, ValueError),
    *mapchars(0x09d8, 0x09db, ValueError),
    ((chr(0x09de),), ValueError),
    *mapchars(0x09e4, 0x09e5, ValueError),
    *mapchars(0x09fb, 0x0a01, ValueError),
    *mapchars(0x0a03, 0x0a04, ValueError),
    *mapchars(0x0a0b, 0x0a0e, ValueError),
    *mapchars(0x0a11, 0x0a12, ValueError),
    ((chr(0x0a29),), ValueError),
    ((chr(0x0a31),), ValueError),
    ((chr(0x0a34),), ValueError),
    ((chr(0x0a37),), ValueError),
    *mapchars(0x0a3a, 0x0a3b, ValueError),
    ((chr(0x0a3d),), ValueError),
    *mapchars(0x0a43, 0x0a46, ValueError),
    *mapchars(0x0a49, 0x0a4a, ValueError),
    *mapchars(0x0a4e, 0x0a58, ValueError),
    ((chr(0x0a5d),), ValueError),
    *mapchars(0x0a5f, 0x0a65, ValueError),
    *mapchars(0x0a75, 0x0a80, ValueError),
    ((chr(0x0a84),), ValueError),
    ((chr(0x0a8c),), ValueError),
    ((chr(0x0a8e),), ValueError),
    ((chr(0x0a92),), ValueError),
    ((chr(0x0aa9),), ValueError),
    ((chr(0x0ab1),), ValueError),
    ((chr(0x0ab4),), ValueError),
    *mapchars(0x0aba, 0x0abb, ValueError),
    ((chr(0x0ac6),), ValueError),
    ((chr(0x0aca),), ValueError),
    *mapchars(0x0ace, 0x0acf, ValueError),
    *mapchars(0x0ad1, 0x0adf, ValueError),
    *mapchars(0x0ae1, 0x0ae5, ValueError),
    *mapchars(0x0af0, 0x0b00, ValueError),
    ((chr(0x0b04),), ValueError),
    *mapchars(0x0b0d, 0x0b0e, ValueError),
    *mapchars(0x0b11, 0x0b12, ValueError),
    ((chr(0x0b29),), ValueError),
    ((chr(0x0b31),), ValueError),
    *mapchars(0x0b34, 0x0b35, ValueError),
    *mapchars(0x0b3a, 0x0b3b, ValueError),
    *mapchars(0x0b44, 0x0b46, ValueError),
    *mapchars(0x0b49, 0x0b4a, ValueError),
    *mapchars(0x0b4e, 0x0b55, ValueError),
    *mapchars(0x0b58, 0x0b5b, ValueError),
    ((chr(0x0b5e),), ValueError),
    *mapchars(0x0b62, 0x0b65, ValueError),
    *mapchars(0x0b71, 0x0b81, ValueError),
    ((chr(0x0b84),), ValueError),
    *mapchars(0x0b8b, 0x0b8d, ValueError),
    ((chr(0x0b91),), ValueError),
    *mapchars(0x0b96, 0x0b98, ValueError),
    ((chr(0x0b9b),), ValueError),
    ((chr(0x0b9d),), ValueError),
    *mapchars(0x0ba0, 0x0ba2, ValueError),
    *mapchars(0x0ba5, 0x0ba7, ValueError),
    *mapchars(0x0bab, 0x0bad, ValueError),
    ((chr(0x0bb6),), ValueError),
    *mapchars(0x0bba, 0x0bbd, ValueError),
    *mapchars(0x0bc3, 0x0bc5, ValueError),
    ((chr(0x0bc9),), ValueError),
    *mapchars(0x0bce, 0x0bd6, ValueError),
    *mapchars(0x0bd8, 0x0be6, ValueError),
    *mapchars(0x0bf3, 0x0c00, ValueError),
    ((chr(0x0c04),), ValueError),
    ((chr(0x0c0d),), ValueError),
    ((chr(0x0c11),), ValueError),
    ((chr(0x0c29),), ValueError),
    ((chr(0x0c34),), ValueError),
    *mapchars(0x0c3a, 0x0c3d, ValueError),
    ((chr(0x0c45),), ValueError),
    ((chr(0x0c49),), ValueError),
    *mapchars(0x0c4e, 0x0c54, ValueError),
    *mapchars(0x0c57, 0x0c5f, ValueError),
    *mapchars(0x0c62, 0x0c65, ValueError),
    *mapchars(0x0c70, 0x0c81, ValueError),
    ((chr(0x0c84),), ValueError),
    ((chr(0x0c8d),), ValueError),
    ((chr(0x0c91),), ValueError),
    ((chr(0x0ca9),), ValueError),
    ((chr(0x0cb4),), ValueError),
    *mapchars(0x0cba, 0x0cbd, ValueError),
    ((chr(0x0cc5),), ValueError),
    ((chr(0x0cc9),), ValueError),
    *mapchars(0x0cce, 0x0cd4, ValueError),
    *mapchars(0x0cd7, 0x0cdd, ValueError),
    ((chr(0x0cdf),), ValueError),
    *mapchars(0x0ce2, 0x0ce5, ValueError),
    *mapchars(0x0cf0, 0x0d01, ValueError),
    ((chr(0x0d04),), ValueError),
    ((chr(0x0d0d),), ValueError),
    ((chr(0x0d11),), ValueError),
    ((chr(0x0d29),), ValueError),
    *mapchars(0x0d3a, 0x0d3d, ValueError),
    *mapchars(0x0d44, 0x0d45, ValueError),
    ((chr(0x0d49),), ValueError),
    *mapchars(0x0d4e, 0x0d56, ValueError),
    *mapchars(0x0d58, 0x0d5f, ValueError),
    *mapchars(0x0d62, 0x0d65, ValueError),
    *mapchars(0x0d70, 0x0d81, ValueError),
    ((chr(0x0d84),), ValueError),
    *mapchars(0x0d97, 0x0d99, ValueError),
    ((chr(0x0db2),), ValueError),
    ((chr(0x0dbc),), ValueError),
    *mapchars(0x0dbe, 0x0dbf, ValueError),
    *mapchars(0x0dc7, 0x0dc9, ValueError),
    *mapchars(0x0dcb, 0x0dce, ValueError),
    ((chr(0x0dd5),), ValueError),
    ((chr(0x0dd7),), ValueError),
    *mapchars(0x0de0, 0x0df1, ValueError),
    *mapchars(0x0df5, 0x0e00, ValueError),
    *mapchars(0x0e3b, 0x0e3e, ValueError),
    *mapchars(0x0e5c, 0x0e80, ValueError),
    ((chr(0x0e83),), ValueError),
    *mapchars(0x0e85, 0x0e86, ValueError),
    ((chr(0x0e89),), ValueError),
    *mapchars(0x0e8b, 0x0e8c, ValueError),
    *mapchars(0x0e8e, 0x0e93, ValueError),
    ((chr(0x0e98),), ValueError),
    ((chr(0x0ea0),), ValueError),
    ((chr(0x0ea4),), ValueError),
    ((chr(0x0ea6),), ValueError),
    *mapchars(0x0ea8, 0x0ea9, ValueError),
    ((chr(0x0eac),), ValueError),
    ((chr(0x0eba),), ValueError),
    *mapchars(0x0ebe, 0x0ebf, ValueError),
    ((chr(0x0ec5),), ValueError),
    ((chr(0x0ec7),), ValueError),
    *mapchars(0x0ece, 0x0ecf, ValueError),
    *mapchars(0x0eda, 0x0edb, ValueError),
    *mapchars(0x0ede, 0x0eff, ValueError),
    ((chr(0x0f48),), ValueError),
    *mapchars(0x0f6b, 0x0f70, ValueError),
    *mapchars(0x0f8c, 0x0f8f, ValueError),
    ((chr(0x0f98),), ValueError),
    ((chr(0x0fbd),), ValueError),
    *mapchars(0x0fcd, 0x0fce, ValueError),
    *mapchars(0x0fd0, 0x0fff, ValueError),
    ((chr(0x1022),), ValueError),
    ((chr(0x1028),), ValueError),
    ((chr(0x102b),), ValueError),
    *mapchars(0x1033, 0x1035, ValueError),
    *mapchars(0x103a, 0x103f, ValueError),
    *mapchars(0x105a, 0x109f, ValueError),
    *mapchars(0x10c6, 0x10cf, ValueError),
    *mapchars(0x10f9, 0x10fa, ValueError),
    *mapchars(0x10fc, 0x10ff, ValueError),
    *mapchars(0x115a, 0x115e, ValueError),
    *mapchars(0x11a3, 0x11a7, ValueError),
    *mapchars(0x11fa, 0x11ff, ValueError),
    ((chr(0x1207),), ValueError),
    ((chr(0x1247),), ValueError),
    ((chr(0x1249),), ValueError),
    *mapchars(0x124e, 0x124f, ValueError),
    ((chr(0x1257),), ValueError),
    ((chr(0x1259),), ValueError),
    *mapchars(0x125e, 0x125f, ValueError),
    ((chr(0x1287),), ValueError),
    ((chr(0x1289),), ValueError),
    *mapchars(0x128e, 0x128f, ValueError),
    ((chr(0x12af),), ValueError),
    ((chr(0x12b1),), ValueError),
    *mapchars(0x12b6, 0x12b7, ValueError),
    ((chr(0x12bf),), ValueError),
    ((chr(0x12c1),), ValueError),
    *mapchars(0x12c6, 0x12c7, ValueError),
    ((chr(0x12cf),), ValueError),
    ((chr(0x12d7),), ValueError),
    ((chr(0x12ef),), ValueError),
    ((chr(0x130f),), ValueError),
    ((chr(0x1311),), ValueError),
    *mapchars(0x1316, 0x1317, ValueError),
    ((chr(0x131f),), ValueError),
    ((chr(0x1347),), ValueError),
    *mapchars(0x135b, 0x1360, ValueError),
    *mapchars(0x137d, 0x139f, ValueError),
    *mapchars(0x13f5, 0x1400, ValueError),
    *mapchars(0x1677, 0x167f, ValueError),
    *mapchars(0x169d, 0x169f, ValueError),
    *mapchars(0x16f1, 0x16ff, ValueError),
    ((chr(0x170d),), ValueError),
    *mapchars(0x1715, 0x171f, ValueError),
    *mapchars(0x1737, 0x173f, ValueError),
    *mapchars(0x1754, 0x175f, ValueError),
    ((chr(0x176d),), ValueError),
    ((chr(0x1771),), ValueError),
    *mapchars(0x1774, 0x177f, ValueError),
    *mapchars(0x17dd, 0x17df, ValueError),
    *mapchars(0x17ea, 0x17ff, ValueError),
    ((chr(0x180f),), ValueError),
    *mapchars(0x181a, 0x181f, ValueError),
    *mapchars(0x1878, 0x187f, ValueError),
    *mapchars(0x18aa, 0x1dff, ValueError),
    *mapchars(0x1e9c, 0x1e9f, ValueError),
    *mapchars(0x1efa, 0x1eff, ValueError),
    *mapchars(0x1f16, 0x1f17, ValueError),
    *mapchars(0x1f1e, 0x1f1f, ValueError),
    *mapchars(0x1f46, 0x1f47, ValueError),
    *mapchars(0x1f4e, 0x1f4f, ValueError),
    ((chr(0x1f58),), ValueError),
    ((chr(0x1f5a),), ValueError),
    ((chr(0x1f5c),), ValueError),
    ((chr(0x1f5e),), ValueError),
    *mapchars(0x1f7e, 0x1f7f, ValueError),
    ((chr(0x1fb5),), ValueError),
    ((chr(0x1fc5),), ValueError),
    *mapchars(0x1fd4, 0x1fd5, ValueError),
    ((chr(0x1fdc),), ValueError),
    *mapchars(0x1ff0, 0x1ff1, ValueError),
    ((chr(0x1ff5),), ValueError),
    ((chr(0x1fff),), ValueError),
    *mapchars(0x2053, 0x2056, ValueError),
    *mapchars(0x2058, 0x205e, ValueError),
    *mapchars(0x2064, 0x2069, ValueError),
    *mapchars(0x2072, 0x2073, ValueError),
    *mapchars(0x208f, 0x209f, ValueError),
    *mapchars(0x20b2, 0x20cf, ValueError),
    *mapchars(0x20eb, 0x20ff, ValueError),
    *mapchars(0x213b, 0x213c, ValueError),
    *mapchars(0x214c, 0x2152, ValueError),
    *mapchars(0x2184, 0x218f, ValueError),
    *mapchars(0x23cf, 0x23ff, ValueError),
    *mapchars(0x2427, 0x243f, ValueError),
    *mapchars(0x244b, 0x245f, ValueError),
    ((chr(0x24ff),), ValueError),
    *mapchars(0x2614, 0x2615, ValueError),
    ((chr(0x2618),), ValueError),
    *mapchars(0x267e, 0x267f, ValueError),
    *mapchars(0x268a, 0x2700, ValueError),
    ((chr(0x2705),), ValueError),
    *mapchars(0x270a, 0x270b, ValueError),
    ((chr(0x2728),), ValueError),
    ((chr(0x274c),), ValueError),
    ((chr(0x274e),), ValueError),
    *mapchars(0x2753, 0x2755, ValueError),
    ((chr(0x2757),), ValueError),
    *mapchars(0x275f, 0x2760, ValueError),
    *mapchars(0x2795, 0x2797, ValueError),
    ((chr(0x27b0),), ValueError),
    *mapchars(0x27bf, 0x27cf, ValueError),
    *mapchars(0x27ec, 0x27ef, ValueError),
    *mapchars(0x2b00, 0x2e7f, ValueError),
    ((chr(0x2e9a),), ValueError),
    *mapchars(0x2ef4, 0x2eff, ValueError),
    *mapchars(0x2fd6, 0x2fef, ValueError),
    *mapchars(0x2ffc, 0x2fff, ValueError),
    ((chr(0x3040),), ValueError),
    *mapchars(0x3097, 0x3098, ValueError),
    *mapchars(0x3100, 0x3104, ValueError),
    *mapchars(0x312d, 0x3130, ValueError),
    ((chr(0x318f),), ValueError),
    *mapchars(0x31b8, 0x31ef, ValueError),
    *mapchars(0x321d, 0x321f, ValueError),
    *mapchars(0x3244, 0x3250, ValueError),
    *mapchars(0x327c, 0x327e, ValueError),
    *mapchars(0x32cc, 0x32cf, ValueError),
    ((chr(0x32ff),), ValueError),
    *mapchars(0x3377, 0x337a, ValueError),
    *mapchars(0x33de, 0x33df, ValueError),
    ((chr(0x33ff),), ValueError),
    *mapchars(0x4db6, 0x4dff, ValueError),
    *mapchars(0x9fa6, 0x9fff, ValueError),
    *mapchars(0xa48d, 0xa48f, ValueError),
    *mapchars(0xa4c7, 0xabff, ValueError),
    *mapchars(0xd7a4, 0xd7ff, ValueError),
    *mapchars(0xfa2e, 0xfa2f, ValueError),
    *mapchars(0xfa6b, 0xfaff, ValueError),
    *mapchars(0xfb07, 0xfb12, ValueError),
    *mapchars(0xfb18, 0xfb1c, ValueError),
    ((chr(0xfb37),), ValueError),
    ((chr(0xfb3d),), ValueError),
    ((chr(0xfb3f),), ValueError),
    ((chr(0xfb42),), ValueError),
    ((chr(0xfb45),), ValueError),
    *mapchars(0xfbb2, 0xfbd2, ValueError),
    *mapchars(0xfd40, 0xfd4f, ValueError),
    *mapchars(0xfd90, 0xfd91, ValueError),
    *mapchars(0xfdc8, 0xfdcf, ValueError),
    *mapchars(0xfdfd, 0xfdff, ValueError),
    *mapchars(0xfe10, 0xfe1f, ValueError),
    *mapchars(0xfe24, 0xfe2f, ValueError),
    *mapchars(0xfe47, 0xfe48, ValueError),
    ((chr(0xfe53),), ValueError),
    ((chr(0xfe67),), ValueError),
    *mapchars(0xfe6c, 0xfe6f, ValueError),
    ((chr(0xfe75),), ValueError),
    *mapchars(0xfefd, 0xfefe, ValueError),
    ((chr(0xff00),), ValueError),
    *mapchars(0xffbf, 0xffc1, ValueError),
    *mapchars(0xffc8, 0xffc9, ValueError),
    *mapchars(0xffd0, 0xffd1, ValueError),
    *mapchars(0xffd8, 0xffd9, ValueError),
    *mapchars(0xffdd, 0xffdf, ValueError),
    ((chr(0xffe7),), ValueError),
    *mapchars(0xffef, 0xfff8, ValueError),
    *mapchars(0x10000, 0x102ff, ValueError),
    ((chr(0x1031f),), ValueError),
    *mapchars(0x10324, 0x1032f, ValueError),
    *mapchars(0x1034b, 0x103ff, ValueError),
    *mapchars(0x10426, 0x10427, ValueError),
    *mapchars(0x1044e, 0x1cfff, ValueError),
    *mapchars(0x1d0f6, 0x1d0ff, ValueError),
    *mapchars(0x1d127, 0x1d129, ValueError),
    *mapchars(0x1d1de, 0x1d3ff, ValueError),
    ((chr(0x1d455),), ValueError),
    ((chr(0x1d49d),), ValueError),
    *mapchars(0x1d4a0, 0x1d4a1, ValueError),
    *mapchars(0x1d4a3, 0x1d4a4, ValueError),
    *mapchars(0x1d4a7, 0x1d4a8, ValueError),
    ((chr(0x1d4ad),), ValueError),
    ((chr(0x1d4ba),), ValueError),
    ((chr(0x1d4bc),), ValueError),
    ((chr(0x1d4c1),), ValueError),
    ((chr(0x1d4c4),), ValueError),
    ((chr(0x1d506),), ValueError),
    *mapchars(0x1d50b, 0x1d50c, ValueError),
    ((chr(0x1d515),), ValueError),
    ((chr(0x1d51d),), ValueError),
    ((chr(0x1d53a),), ValueError),
    ((chr(0x1d53f),), ValueError),
    ((chr(0x1d545),), ValueError),
    *mapchars(0x1d547, 0x1d549, ValueError),
    ((chr(0x1d551),), ValueError),
    *mapchars(0x1d6a4, 0x1d6a7, ValueError),
    *mapchars(0x1d7ca, 0x1d7cd, ValueError),
    *mapchars(0x1d800, 0x1fffd, ValueError),
    *mapchars(0x2a6d7, 0x2f7ff, ValueError),
    *mapchars(0x2fa1e, 0x2fffd, ValueError),
    *mapchars(0x30000, 0x3fffd, ValueError),
    *mapchars(0x40000, 0x4fffd, ValueError),
    *mapchars(0x50000, 0x5fffd, ValueError),
    *mapchars(0x60000, 0x6fffd, ValueError),
    *mapchars(0x70000, 0x7fffd, ValueError),
    *mapchars(0x80000, 0x8fffd, ValueError),
    *mapchars(0x90000, 0x9fffd, ValueError),
    *mapchars(0xa0000, 0xafffd, ValueError),
    *mapchars(0xb0000, 0xbfffd, ValueError),
    *mapchars(0xc0000, 0xcfffd, ValueError),
    *mapchars(0xd0000, 0xdfffd, ValueError),
    ((chr(0xe0000),), ValueError),
    *mapchars(0xe0002, 0xe001f, ValueError),
    *mapchars(0xe0080, 0xefffd, ValueError),

    # Non-ASCII space
    (('\u0020',), ' '),
    (('\u00A0',), ' '),
    (('\u1680',), ' '),
    *mapchars(0x2000, 0x200a, ' '),
    (('\u202f',), ' '),
    (('\u205f',), ' '),
    (('\u3000',), ' '),

    # Mapped to nothing
    (('\u00ad',), ''),
    (('\u034f',), ''),
    (('\u1806',), ''),
    *mapchars(0x180b, 0x180d, ''),
    *mapchars(0x200b, 0x200d, ''),
    (('\u2060',), ''),
    *mapchars(0xfe00, 0xfe0f, ''),
    (('\ufeff',), ''),

    # ASCII control characters
    *mapchars(0x00, 0x1f, ValueError),
    (('\007f',), ValueError),

    # Non-ASCII control characters
    *mapchars(0x80, 0x9f, ValueError),
    (('\u06dd',), ValueError),
    (('\u070f',), ValueError),
    (('\u180e',), ValueError),
    (('\u200c',), ''),
    (('\u200d',), ''),
    (('\u2028',), ValueError),
    (('\u2029',), ValueError),
    (('\u2060',), ''),
    *mapchars(0x2061, 0x2063, ValueError),
    *mapchars(0x206a, 0x206f, ValueError),
    (('\ufeff',), ''),
    *mapchars(0xfff9, 0xfffc, ValueError),
    *mapchars(0x1d173, 0x1d17a, ValueError),

    # Private Use characters
    *mapchars(0xe000, 0xf8ff, ValueError),
    *mapchars(0xf0000, 0xffffd, ValueError),
    *mapchars(0x100000, 0x10fffd, ValueError),

    # Non-character code points
    *mapchars(0xfdd0, 0xfdef, ValueError),
    *mapchars(0x1fffe, 0x1ffff, ValueError),
    *mapchars(0x2fffe, 0x2ffff, ValueError),
    *mapchars(0x3fffe, 0x3ffff, ValueError),
    *mapchars(0x4fffe, 0x4ffff, ValueError),
    *mapchars(0x5fffe, 0x5ffff, ValueError),
    *mapchars(0x6fffe, 0x6ffff, ValueError),
    *mapchars(0x7fffe, 0x7ffff, ValueError),
    *mapchars(0x8fffe, 0x8ffff, ValueError),
    *mapchars(0x9fffe, 0x9ffff, ValueError),
    *mapchars(0xafffe, 0xaffff, ValueError),
    *mapchars(0xbfffe, 0xbffff, ValueError),
    *mapchars(0xcfffe, 0xcffff, ValueError),
    *mapchars(0xdfffe, 0xdffff, ValueError),
    *mapchars(0xefffe, 0xeffff, ValueError),
    *mapchars(0xffffe, 0xfffff, ValueError),
    *mapchars(0x10fffe, 0x10ffff, ValueError),

    # Surrogate code points
    *mapchars(0xd800, 0xdfff, ValueError),

    # Inappropriate for plain text characters
    *mapchars(0xfff9, 0xfffd, ValueError),

    # Inappropriate for canonical representation characters
    *mapchars(0x2ff0, 0x2ffb, ValueError),

    # Change display properties or deprecated characters
    *mapchars(0x0340, 0x0341, ValueError),
    *mapchars(0x200e, 0x200f, ValueError),
    *mapchars(0x202a, 0x202e, ValueError),
    *mapchars(0x206a, 0x206f, ValueError),

    # Tagging characters
    (chr(0xe0001), ValueError),
    *mapchars(0xe0020, 0xe007f, ValueError),

    # Bi-directional characters
    (('\u0627\u0031',), ValueError),
    (('\u0627\u0031\u0628',), '\u0627\u0031\u0628'),

    # Examples from RFC 4013
    (('I\u00adX',), 'IX'),
    (('user',), 'user'),
    (('USER',), 'USER'),
    (('\u00aa',), 'a'),
    (('\u2168',), 'IX'),
    (('\u0007',), ValueError),
    (('\u0627\u0031',), ValueError)
)


#
# Tests
#

class TestBaseAuth(unittest.TestCase):
    test_prepare = unittest.skipIf(os.getenv('QUICKTEST'), 'slow')(
        makerunner(BaseAuth.prepare, STRINGS)
    )


#
# Boilerplate
#

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Tests for the ManageSieve API."""

#
# Copyright 2024  Odin Kroeger
#
# This file is part of SieveManager.
#
# SieveManager is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# SieveManager is distributed in the hope that it will be useful,
# but WITHOUT ALL WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SieveManager. If not, see <https://www.gnu.org/licenses/>.
#

# pylint: disable=missing-class-docstring,missing-function-docstring


#
# Modules
#

import os
import pathlib
import re
import sys
import unittest

from typing import Final

sys.path.append(os.path.realpath(pathlib.Path(__file__).parents[1]))

from sievemgr import (Atom, Capabilities, Response, SieveManager,
                      SieveError, SieveConnectionError,
                      SieveOperationError, SieveProtocolError, URL)
from . import Test, makerunner


#
# Globals
#

SCRIPTNAMES: Final[tuple[Test[bool], ...]] = (
    # Simple tests
    (('',), False),
    *[((chr(i),), False) for i in range(32)],
    (('\0x7f',), False),
    *[((chr(i),), False) for i in range(0x80, 0xa0)],
    (('\0x2028',), False),
    (('\0x2029',), False),
    (('\0xfeff',), False),
    (('\n',), False),
    (('\r',), False),
    ('\0', False),

    # Paradigmatic cases
    (('foo',), True),
    (('bar.sieve',), True),

    # Edge cases
    ((' ',), True),

    # Non-ASCII characters
    (('𝚏õȫ',), True),
    (('𝕗ōȏ.ⓢℹéṽ𝕖',), True)
)


SCRIPTNAME_ERRORS: Final[tuple[Test[bool], ...]] = tuple(
    ((*args, True), exp if exp else ValueError) for args, exp in SCRIPTNAMES
)


CAPABILITIES: Final[tuple[Test[Capabilities], ...]] = (
    # Simple cases
    (([],), Capabilities()),
    (([['IMPLEMENTATION', '']],), Capabilities(implementation='')),
    (([['IMPLEMENTATION', 'Foo v1']],), Capabilities(implementation='Foo v1')),
    (([['SASL', '']],), Capabilities()),
    (([['SASL', 'FOO']],), Capabilities(sasl=('foo',))),
    (([['SASL', 'FOO BAR']],), Capabilities(sasl=('foo', 'bar'))),
    (([['SIEVE', '']],), Capabilities()),
    (([['SIEVE', 'foo']],), Capabilities(sieve=('foo',))),
    (([['SIEVE', 'foo bar']],), Capabilities(sieve=('foo', 'bar'))),
    (([['STARTTLS']],), Capabilities(starttls=True)),
    (([['MAXREDIRECTS', '0']],), Capabilities(maxredirects=0)),
    (([['MAXREDIRECTS', '1']],), Capabilities(maxredirects=1)),
    (([['MAXREDIRECTS', '1']],), Capabilities(maxredirects=1)),
    (([['MAXREDIRECTS', '4294967295']],),
     Capabilities(maxredirects=4294967295)),
    (([['NOTIFY', '']],), Capabilities()),
    (([['NOTIFY', 'foo']],), Capabilities(notify=('foo',))),
    (([['NOTIFY', 'foo bar']],), Capabilities(notify=('foo', 'bar'))),
    (([['OWNER', 'user']],), Capabilities(owner='user')),
    (([['VERSION', '1.0']],), Capabilities(version='1.0')),
    (([['UNAUTHENTICATE']],), Capabilities(unauthenticate=True)),

    # Examples from RFC 5804
    (
        ([
            ['IMPlemENTATION', 'Example1 ManageSieved v001'],
            ['SASl', 'DIGEST-MD5 GSSAPI'],
            ['SIeVE', 'fileinto vacation'],
            ['StaRTTLS'],
            ['NOTIFY', 'xmpp mailto'],
            ['MAXREdIRECTS', '5'],
            ['VERSION', '1.0']
        ],),
        Capabilities(
             implementation='Example1 ManageSieved v001',
             sasl=('digest-md5', 'gssapi'),
             sieve=('fileinto', 'vacation'),
             starttls=True,
             notify=('xmpp', 'mailto'),
             maxredirects=5,
             version='1.0'
        )
    ),
    (
        ([
            ['IMPlemENTATION', 'Example1 ManageSieved v001'],
            ['SASl', 'DIGEST-MD5 GSSAPI'],
            ['SIeVE', 'fileinto vacation'],
            ['StaRTTLS'],
            ['NOTIFY', 'xmpp mailto'],
            ['MAXREdIRECTS', '5'],
            ['OWNER', 'alexey@example.com'],
            ['VERSION', '1.0']
        ],),
        Capabilities(
            implementation='Example1 ManageSieved v001',
            sasl=('digest-md5', 'gssapi'),
            sieve=('fileinto', 'vacation'),
            starttls=True,
            notify=('xmpp', 'mailto'),
            owner='alexey@example.com',
            maxredirects=5,
            version='1.0'
        )
    ),

    # Real-world data
    (
        ([
            ['IMPLEMENTATION', 'Dovecot Pigeonhole'],
            ['SIEVE', 'fileinto reject envelope encoded-character '
                      'vacation subaddress comparator-i;ascii-numeric '
                      'relational regex imap4flags copy include '
                      'variables body enotify environment '
                      'mailbox date index ihave duplicate '
                      'vacation-seconds imapflags notify '
                      'vnd.dovecot.pgp-encrypt editheader'],
            ['NOTIFY', 'mailto'],
            ['SASL', ''],
            ['STARTTLS'],
            ['VERSION', '1.0']
        ],),
        Capabilities(
            implementation='Dovecot Pigeonhole',
            sieve=('fileinto', 'reject', 'envelope', 'encoded-character',
                   'vacation', 'subaddress', 'comparator-i;ascii-numeric',
                   'relational', 'regex', 'imap4flags', 'copy', 'include',
                   'variables', 'body', 'enotify', 'environment',
                   'mailbox', 'date', 'index', 'ihave', 'duplicate',
                   'vacation-seconds', 'imapflags', 'notify',
                   'vnd.dovecot.pgp-encrypt', 'editheader'),
            notify=('mailto',),
            starttls=True,
            version='1.0'
        )
    ),
    (
        ([
            ['IMPLEMENTATION', 'Dovecot Pigeonhole'],
            ['SIEVE', 'fileinto reject envelope encoded-character '
                      'vacation subaddress comparator-i;ascii-numeric '
                      'relational regex imap4flags copy include '
                      'variables body enotify environment '
                      'mailbox date index ihave duplicate '
                      'vacation-seconds imapflags notify '
                      'vnd.dovecot.pgp-encrypt editheader'],
            ['NOTIFY', 'mailto'],
            ['SASL', 'PLAIN LOGIN OAUTHBEARER XOAUTH2'],
            ['VERSION', '1.0']
        ],),
        Capabilities(
            implementation='Dovecot Pigeonhole',
            sieve=('fileinto', 'reject', 'envelope', 'encoded-character',
                   'vacation', 'subaddress', 'comparator-i;ascii-numeric',
                   'relational', 'regex', 'imap4flags', 'copy', 'include',
                   'variables', 'body', 'enotify', 'environment',
                   'mailbox', 'date', 'index', 'ihave', 'duplicate',
                   'vacation-seconds', 'imapflags', 'notify',
                   'vnd.dovecot.pgp-encrypt', 'editheader'),
            notify=('mailto',),
            sasl=('plain', 'login', 'oauthbearer', 'xoauth2'),
            version='1.0'
        )
    ),
    (
        ([
            ['IMPLEMENTATION', 'Cyrus timsieved (Murder) 2.5.17'],
            ['SASL', 'PLAIN'],
            ['SIEVE', 'comparator-i;ascii-numeric fileinto reject '
                      'vacation vacation-seconds imapflags notify '
                      'envelope imap4flags relational regex '
                      'subaddress copy date'],
            ['UNAUTHENTICATE']
        ],),
        Capabilities(
            implementation='Cyrus timsieved (Murder) 2.5.17',
            sasl=('plain',),
            sieve=('comparator-i;ascii-numeric', 'fileinto', 'reject',
                   'vacation', 'vacation-seconds', 'imapflags', 'notify',
                   'envelope', 'imap4flags', 'relational', 'regex',
                   'subaddress', 'copy', 'date'),
            unauthenticate=True
        )
    ),

    # Errors
    (([[]],), SieveProtocolError),
    (([['IMPLEMENTATION']],), SieveProtocolError),
    (([['IMPLEMENTATION', None]],), SieveProtocolError),
    (([['IMPLEMENTATION', True]],), SieveProtocolError),
    (([['IMPLEMENTATION', 0]],), SieveProtocolError),
    (([['IMPLEMENTATION', 1]],), SieveProtocolError),
    (([['IMPLEMENTATION', []]],), SieveProtocolError),
    (([['IMPLEMENTATION', 'foo', 'bar']],), SieveProtocolError),
    (([['SASL']],), SieveProtocolError),
    (([['SASL', None]],), SieveProtocolError),
    (([['SASL', True]],), SieveProtocolError),
    (([['SASL', 0]],), SieveProtocolError),
    (([['SASL', 1]],), SieveProtocolError),
    (([['SASL', []]],), SieveProtocolError),
    (([['SASL', 'foo', 'bar']],), SieveProtocolError),
    (([['SIEVE']],), SieveProtocolError),
    (([['SIEVE', None]],), SieveProtocolError),
    (([['SIEVE', True]],), SieveProtocolError),
    (([['SIEVE', 0]],), SieveProtocolError),
    (([['SIEVE', 1]],), SieveProtocolError),
    (([['SIEVE', []]],), SieveProtocolError),
    (([['SIEVE', 'foo', 'bar']],), SieveProtocolError),
    (([['MAXREDIRECTS']],), SieveProtocolError),
    (([['MAXREDIRECTS', None]],), SieveProtocolError),
    (([['MAXREDIRECTS', True]],), SieveProtocolError),
    (([['MAXREDIRECTS', 0]],), SieveProtocolError),
    (([['MAXREDIRECTS', 1]],), SieveProtocolError),
    (([['MAXREDIRECTS', []]],), SieveProtocolError),
    (([['MAXREDIRECTS', 'foo', 'bar']],), SieveProtocolError),
    (([['NOTIFY']],), SieveProtocolError),
    (([['NOTIFY', None]],), SieveProtocolError),
    (([['NOTIFY', True]],), SieveProtocolError),
    (([['NOTIFY', 0]],), SieveProtocolError),
    (([['NOTIFY', 1]],), SieveProtocolError),
    (([['NOTIFY', []]],), SieveProtocolError),
    (([['NOTIFY', 'foo', 'bar']],), SieveProtocolError),
    (([['OWNER']],), SieveProtocolError),
    (([['OWNER', None]],), SieveProtocolError),
    (([['OWNER', True]],), SieveProtocolError),
    (([['OWNER', 0]],), SieveProtocolError),
    (([['OWNER', 1]],), SieveProtocolError),
    (([['OWNER', []]],), SieveProtocolError),
    (([['OWNER', 'foo', 'bar']],), SieveProtocolError),
    (([['VERSION']],), SieveProtocolError),
    (([['VERSION', None]],), SieveProtocolError),
    (([['VERSION', True]],), SieveProtocolError),
    (([['VERSION', 0]],), SieveProtocolError),
    (([['VERSION', 1]],), SieveProtocolError),
    (([['VERSION', []]],), SieveProtocolError),
    (([['VERSION', 'foo', 'bar']],), SieveProtocolError),
)


MESSAGES: Final[tuple[Test[str], ...]] = (
    # No message
    ((Response(Atom('OK')),), 'server says OK'),
    ((Response(Atom('NO')),), 'server says NO'),
    ((Response(Atom('BYE')),), 'server says BYE'),

    # Empty message
    ((Response(Atom('OK'), message=''),), 'server says OK'),
    ((Response(Atom('NO'), message=''),), 'server says NO'),
    ((Response(Atom('BYE'), message=''),), 'server says BYE'),

    # Message
    ((Response(Atom('OK'), message='foo'),), 'foo'),
    ((Response(Atom('NO'), message='foo'),), 'foo'),
    ((Response(Atom('BYE'), message='foo'),), 'foo')
)


RESPONSES: Final[tuple[Test[Response], ...]] = (
    # Simple cases
    (([Atom('OK')],), Response(response=Atom('OK'))),
    (([Atom('NO')],), Response(response=Atom('NO'))),
    (([Atom('BYE')],), Response(response=Atom('BYE'))),

    # Response codes
    (([Atom('OK'), []],),
     Response(response=Atom('OK'))),
    (([Atom('NO'), []],),
     Response(response=Atom('NO'))),
    (([Atom('BYE'), []],),
     Response(response=Atom('BYE'))),
    (([Atom('OK'), [Atom('foo')]],),
     Response(response=Atom('OK'), code=(Atom('foo'),))),
    (([Atom('NO'), [Atom('foo')]],),
     Response(response=Atom('NO'), code=(Atom('foo'),))),
    (([Atom('BYE'), [Atom('foo')]],),
     Response(response=Atom('BYE'), code=(Atom('foo'),))),
    (([Atom('OK'), ['foo', 'bar']],),
     Response(response=Atom('OK'), code=('foo', 'bar'))),
    (([Atom('NO'), ['foo', 'bar']],),
     Response(response=Atom('NO'), code=('foo', 'bar'))),
    (([Atom('BYE'), ['foo', 'bar']],),
     Response(response=Atom('BYE'), code=('foo', 'bar'))),
    (([Atom('NO'), [None, True, False, 0, 1, [], 'foo']],),
     Response(response=Atom('NO'), code=(None, True, False, 0, 1, [], 'foo'))),
    (([Atom('NO'), [[[]]]],),
     Response(response=Atom('NO'), code=([[]],))),
    (([Atom('NO'), [[[Atom('NIL')]]]],),
     Response(response=Atom('NO'), code=([[Atom('NIL')]],))),

    # Messages
    (([Atom('OK'), 'Success'],),
     Response(response=Atom('OK'), message='Success')),
    (([Atom('NO'), 'Failure'],),
     Response(response=Atom('NO'), message='Failure')),
    (([Atom('BYE'), 'Bye'],),
     Response(response=Atom('BYE'), message='Bye')),

    # Full responses
    (([Atom('OK'), [Atom('WARNINGS')], 'foo'],),
     Response(Atom('OK'), ('warnings',), 'foo')),
    (([Atom('NO'), [Atom('AUTH-TOO-WEAK')], 'Authentication failed'],),
     Response(Atom('NO'), ('auth-too-weak',), 'Authentication failed')),
    (([Atom('NO'), [Atom('ENCRYPT-NEEDED')], 'Authentication failed'],),
     Response(Atom('NO'), ('encrypt-needed',), 'Authentication failed')),
    (([Atom('NO'), [Atom('QUOTA')], 'Over quota'],),
     Response(Atom('NO'), ('quota',), 'Over quota')),
    (([Atom('NO'), [Atom('QUOTA/MAXSCRIPTS')], 'Too many scripts'],),
     Response(Atom('NO'), ('quota/maxscripts',), 'Too many scripts')),
    (([Atom('NO'), [Atom('QUOTA/MAXSIZE')], 'Script too large'],),
     Response(Atom('NO'), ('quota/maxsize',), 'Script too large')),

    # Examples from RFC 5804
    (
        ([
            Atom('OK'), [
                Atom('SASL'),
                'cnNwYXV0aD1lYTQwZjYwMzM1YzQyN2I1NTI3Yjg0ZGJhYmNkZmZmZA=='
            ]
        ],),
        Response(
            Atom('OK'), (
                Atom('SASL'),
                'cnNwYXV0aD1lYTQwZjYwMzM1YzQyN2I1NTI3Yjg0ZGJhYmNkZmZmZA=='
            )
        )
    ),
    (
        ([
            Atom('OK'),
            [Atom('WARNINGS')],
            'line 8: server redirect action limit is 2, '
            'this redirect might be ignored'
        ],),
        Response(
            Atom('OK'),
            (Atom('WARNINGS'),),
            'line 8: server redirect action limit is 2, '
            'this redirect might be ignored'
        )
    ),
    (
        ([Atom('OK'), 'NOOP completed'],),
        Response(Atom('OK'), message='NOOP completed')
    ),
    (
        ([Atom('OK'), [Atom('TAG'), 'STARTTLS-SYNC-42'], 'Done'],),
        Response(Atom('OK'), (Atom('TAG'), 'STARTTLS-SYNC-42'), 'Done')
    ),
    (
        ([Atom('NO'), [Atom('QUOTA/MAXSIZE')], 'Quota exceeded'],),
        Response(Atom('NO'), (Atom('QUOTA/MAXSIZE'),), 'Quota exceeded')
    ),
    (
        ([Atom('NO'), 'line 2: Syntax error'],),
        Response(Atom('NO'), message='line 2: Syntax error')
    ),
    (
        ([Atom('BYE'), 'Too many failed authentication attempts'],),
        Response(
            Atom('BYE'),
            message='Too many failed authentication attempts'
        )
    ),

    # Errors
    (([],), SieveProtocolError),
    (([None],), SieveProtocolError),
    (([[]],), SieveProtocolError),
    ((['foo'],), SieveProtocolError),
    (([Atom('FOO'), None],), SieveProtocolError),
    (([Atom('FOO'), Atom('BAR')],), SieveProtocolError),
    (([Atom('FOO'), [Atom('BAR')], None],), SieveProtocolError),
    (([Atom('FOO'), 'foo', None],), SieveProtocolError),
    (([Atom('FOO'), [Atom('BAR')], 'foo', None],), SieveProtocolError),
    (([Atom('FOO'), [Atom('BAR')], 'foo', Atom('FOO')],), SieveProtocolError),
    (([Atom('FOO'), [Atom('BAR')], 'foo', []],), SieveProtocolError),
    (([Atom('FOO'), [Atom('BAR')], 'foo', 'foo'],), SieveProtocolError)
)


CODES: Final[tuple[Test[bool], ...]] = (
    # Edge cases
    ((Response(Atom('OK')),), False),
    ((Response(Atom('OK')), ''), False),
    ((Response(Atom('OK'), ('',)),), False),
    ((Response(Atom('OK'), ('',)), ''), True),
    ((Response(Atom('OK'), ('',)), 'foo'), False),
    ((Response(Atom('OK')), '/'), False),
    ((Response(Atom('OK'), ('/',)),), False),
    ((Response(Atom('OK'), ('/',)), '/'), True),
    ((Response(Atom('OK'), ('/',)), 'foo'), False),
    ((Response(Atom('OK'), ('',)), '/'), True),
    ((Response(Atom('OK'), ('/',)), ''), True),
    ((Response(Atom('OK'), ('foo',)), ''), False),
    ((Response(Atom('OK'), ('foo',)), '/'), False),

    # Category in tree
    ((Response(Atom('OK'), ('foo',)), 'foo'), True),
    ((Response(Atom('OK'), ('foo/',)), 'foo'), True),
    ((Response(Atom('OK'), ('foo/bar',)), 'foo'), True),
    ((Response(Atom('OK'), ('',)), 'Foo'), False),
    ((Response(Atom('OK'), ('Foo',)), ''), False),
    ((Response(Atom('OK'), ('Foo',)), 'foo'), True),
    ((Response(Atom('OK'), ('FoO/',)), 'foo'), True),
    ((Response(Atom('OK'), ('foO/BAR',)), 'foo'), True),
    ((Response(Atom('OK'), ('',)), 'FOO'), False),
    ((Response(Atom('OK'), ('foo',)), ''), False),
    ((Response(Atom('OK'), ('foo',)), 'FOO'), True),
    ((Response(Atom('OK'), ('foo/',)), 'Foo'), True),
    ((Response(Atom('OK'), ('foo/bar',)), 'fOo'), True),
    ((Response(Atom('OK'), ('foo',)), 'bar'), False),
    ((Response(Atom('OK'), ('foo/',)), 'bar'), False),
    ((Response(Atom('OK'), ('foo/bar',)), 'bar'), False),
    ((Response(Atom('OK'), ('Foo',)), 'bar'), False),
    ((Response(Atom('OK'), ('FoO/',)), 'bar'), False),
    ((Response(Atom('OK'), ('foO/BAR',)), 'bar'), False),
    ((Response(Atom('OK'), ('foo',)), 'Bar'), False),
    ((Response(Atom('OK'), ('foo/',)), 'bAr'), False),
    ((Response(Atom('OK'), ('foo/bar',)), 'baR'), False),

    # Tree in category
    ((Response(Atom('OK'), ('foo',)), ''), False),
    ((Response(Atom('OK'), ('foo',)), '/foo'), False),
    ((Response(Atom('OK'), ('foo',)), '/foo/'), False),
    ((Response(Atom('OK'), ('foo',)), '/foo/bar'), False),
    ((Response(Atom('OK'), ('foo',)), 'foo'), True),
    ((Response(Atom('OK'), ('foo',)), 'foo/'), True),
    ((Response(Atom('OK'), ('foo',)), 'foo/bar'), False),
    ((Response(Atom('OK'), ('Foo',)), 'foo'), True),
    ((Response(Atom('OK'), ('fOo',)), 'foo/'), True),
    ((Response(Atom('OK'), ('foO',)), 'foo/bar'), False),
    ((Response(Atom('OK'), ('foo',)), '/FOO'), False),
    ((Response(Atom('OK'), ('foo',)), '/foO/'), False),
    ((Response(Atom('OK'), ('foo',)), '/foo/BaR'), False),
    ((Response(Atom('OK'), ('foo',)), 'Foo'), True),
    ((Response(Atom('OK'), ('foo',)), 'fOo/'), True),
    ((Response(Atom('OK'), ('foo',)), 'foO/bAR'), False),
    ((Response(Atom('OK'), ('foo',)), 'bar/foo'), False),
    ((Response(Atom('OK'), ('foo',)), 'bar/foo'), False),
    ((Response(Atom('OK'), ('foo',)), 'bar/foo'), False),
    ((Response(Atom('OK'), ('Foo',)), 'bar/foo'), False),
    ((Response(Atom('OK'), ('FoO',)), 'bar/foo'), False),
    ((Response(Atom('OK'), ('foO',)), 'bar/foo'), False),
    ((Response(Atom('OK'), ('foo',)), 'Bar/FOO'), False),
    ((Response(Atom('OK'), ('foo',)), 'bAr/foO'), False),
    ((Response(Atom('OK'), ('foo',)), 'baR/Foo'), False),

    # Tree in tree
    ((Response(Atom('OK'), ('/bar',)), '/bar'), True),
    ((Response(Atom('OK'), ('/bar/',)), '/bar'), True),
    ((Response(Atom('OK'), ('/bar',)), '/bar/'), True),
    ((Response(Atom('OK'), ('/bar',)), '/foo'), False),
    ((Response(Atom('OK'), ('/bar/baz',)), '/bar'), True),
    ((Response(Atom('OK'), ('/bar',)), '/bar/baz'), False),
    ((Response(Atom('OK'), ('foo/bar',)), 'foo/bar'), True),
    ((Response(Atom('OK'), ('foo/bar/',)), 'foo/bar'), True),
    ((Response(Atom('OK'), ('foo/bar',)), 'foo/bar/'), True),
    ((Response(Atom('OK'), ('foo/bar',)), 'bar/foo'), False),
    ((Response(Atom('OK'), ('foo/bar/baz',)), 'foo/bar'), True),
    ((Response(Atom('OK'), ('foo/bar',)), 'foo/bar/baz'), False),

    # Regular expressions
    ((Response(Atom('OK'), ('.',)), '.'), True),
    ((Response(Atom('OK'), ('foo',)), '.'), False),
    ((Response(Atom('OK'), ('foo',)), '.*'), False),
)


ERRORS: Final[tuple[Test[SieveError], ...]] = (
    ((Response(Atom('')),), SieveOperationError(Atom(''))),
    ((Response(Atom('FOO')),), SieveOperationError(Atom('FOO'))),
    ((Response(Atom('OK')),), SieveOperationError(Atom('OK'))),
    ((Response(Atom('NO')),), SieveOperationError()),
    ((Response(Atom('BYE')),), SieveConnectionError()),
)


URLS: Final[tuple[Test[URL], ...]] = (
    (('foo',),
     URL(hostname='foo')),
    (('foo:0',),
     URL(hostname='foo', port=0)),
    (('foo/user',),
     URL(hostname='foo', owner='user')),
    (('foo/user/script',),
     URL(hostname='foo', owner='user', scriptname='script')),
    (('foo/user/script/script',),
     URL(hostname='foo', owner='user', scriptname='script/script')),
    (('sieve://foo',),
     URL(hostname='foo')),
    (('http://foo',),
     URL(scheme='http', hostname='foo')),
    (('user@foo',),
     URL(username='user', hostname='foo')),
    (('user:password@foo',),
     URL(username='user', password='password',  # nosec B106
         hostname='foo')),
    (('user:password@foo/jdoe',),
     URL(username='user', password='password',  # nosec B106
         hostname='foo', owner='jdoe')),
    (('sieve://user:password@foo/jdoe',),
     URL(username='user', password='password',  # nosec B106
         hostname='foo', owner='jdoe')),
    (('sieve://user:password@foo:123/jdoe/script',),
     URL(username='user', password='password',  # nosec B106
         hostname='foo', port=123, owner='jdoe', scriptname='script')),

    # Errors
    (('',), ValueError),
    (('sieve://',), ValueError),
    (('sieve://',), ValueError),
    ((':80',), ValueError),
    (('/foo',), ValueError),
    (('sieve://foo?bar=true',), ValueError),
)


REVERSEURLS: Final[tuple[Test[str], ...]] = tuple(
    ((u,), s if re.match(r'\w+://', s) else 'sieve://' + s)
    for ((s,), u) in URLS
    if not (isinstance(u, type) and issubclass(u, Exception))  # type: ignore
)


#
# Tests
#

class TestSieveManager(unittest.TestCase):
    test_validname = makerunner(SieveManager.validname, SCRIPTNAMES)
    test_validname_chk = makerunner(SieveManager.validname, SCRIPTNAME_ERRORS)


class TestCapabilities(unittest.TestCase):
    test_fromlines = makerunner(Capabilities.fromlines, CAPABILITIES)


class TestResponse(unittest.TestCase):
    test_str = makerunner(Response.__str__, MESSAGES)
    test_fromline = makerunner(Response.fromline, RESPONSES)
    test_matches = makerunner(Response.matches, CODES)
    test_toerror = makerunner(Response.toerror, ERRORS)


class TestURL(unittest.TestCase):
    test_fromurl = makerunner(URL.fromstr, URLS)
    test_str = makerunner(URL.__str__, REVERSEURLS)


#
# Boilerplate
#

if __name__ == '__main__':
    unittest.main()



"""Test :class:`sievemgr.TermIO`."""

#
# Copyright 2024  Odin Kroeger
#
# This file is part of SieveManager.
#
# SieveManager is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# SieveManager is distributed in the hope that it will be useful,
# but WITHOUT ALL WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SieveManager. If not, see <https://www.gnu.org/licenses/>.
#

# pylint: disable=missing-class-docstring,missing-function-docstring


#
# Modules
#

import os
import pathlib
import sys
import unittest

sys.path.append(os.path.realpath(pathlib.Path(__file__).parents[1]))

from sievemgr import TermIO


#
# Tests
#

class TestTermIO(unittest.TestCase):
    """Tests for :class:`TermIO`."""

    def setUp(self):
        self.tty = TermIO()

    def tearDown(self):
        if not self.tty.closed:
            self.tty.close()

    def test_close(self):
        tty = TermIO()
        tty.close()
        self.assertTrue(tty.closed)

    def test_closed(self):
        self.assertFalse(self.tty.closed)

    def test_fileno(self):
        self.assertTrue(self.tty.fileno() > 0)

    def test_flush(self):
        self.tty.flush()

    def test_isatty(self):
        self.assertTrue(self.tty.isatty())

    def test_read(self):
        self.assertEqual(self.tty.read(0), '')

    def test_readable(self):
        self.assertTrue(self.tty.readable())

    def test_write(self):
        self.tty.write('')

    def test_writable(self):
        return self.assertTrue(self.tty.writable)


#
# Boilerplate
#

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Tests for :class:`sievemgr.BaseACAPConn`."""

#
# Copyright 2024  Odin Kroeger
#
# This file is part of SieveManager.
#
# SieveManager is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# SieveManager is distributed in the hope that it will be useful,
# but WITHOUT ALL WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SieveManager. If not, see <https://www.gnu.org/licenses/>.
#

# pylint: disable=missing-class-docstring,missing-function-docstring


#
# Modules
#

import copy
import io
import os
import pathlib
import sys
import unittest

from typing import Final, IO, Union, Sequence, TypeVar

sys.path.append(os.path.realpath(pathlib.Path(__file__).parents[1]))

from sievemgr import Atom, Line, Word, BaseACAPConn


#
# Types
#

T = TypeVar('T')

MyTest = tuple[bytes, list[Union[Word, io.BytesIO, io.StringIO, bytes]]]


#
# Globals
#

LITERAL: Final[bytes] = b'x' * 1025


BYTESTOLINE: Final[tuple[tuple[bytes, Line], ...]] = (
    # Simple tests
    (b'NIL\r\n', [None]),
    (b'FOO\r\n', [Atom('FOO')]),
    (b'0\r\n', [0]),
    (b'1\r\n', [1]),
    (b'4294967295\r\n', [4_294_967_295]),
    (b'""\r\n', ['']),
    (b'"0"\r\n', ['0']),
    (b'"foo"\r\n', ['foo']),
    (b'{%d+}\r\n%s\r\n' % (len(LITERAL), LITERAL), [LITERAL.decode('utf8')]),
    (b'()\r\n', [[]]),
    (b'(0 1 4294967295)\r\n', [[0, 1, 4_294_967_295]]),

    # Nested lists
    (b'1 (2 3) 4\r\n', [1, [2, 3], 4]),
    (b'1 (2 () 3) 4\r\n', [1, [2, [], 3], 4]),

    # Control characters
    (b'{1+}\r\n\0\r\n', ['\0']),
    (b'{1+}\r\n\r\r\n', ['\r']),
    (b'{1+}\r\n\n\r\n', ['\n']),
    (b'{2+}\r\n\0\r\r\n', ['\0\r']),
    (b'{2+}\r\n\n\r\r\n', ['\n\r']),
    (b'{3+}\r\n\n\0\r\r\n', ['\n\0\r']),

    # Spaces
    (b'" "\r\n', [' ']),
    (b'" foo"\r\n', [' foo']),
    (b'"foo "\r\n', ['foo ']),
    (b'"foo bar"\r\n', ['foo bar']),

    # ManageSieve commands
    (b'Authenticate "PLAIN" "QJIrweAPyo6Q1T9xu"\r\n',
     [Atom('Authenticate'), 'PLAIN', 'QJIrweAPyo6Q1T9xu']),
    (b'StartTls\r\n', [Atom('StartTls')]),
    (b'Logout\r\n', [Atom('Logout')]),
    (b'CAPABILITY\r\n', [Atom('CAPABILITY')]),
    (b'HAVESPACE "myscript" 999999\r\n',
     [Atom('HAVESPACE'), 'myscript', 999999]),
    (b'HAVESPACE "foobar" 435\r\n',
     [Atom('HAVESPACE'), 'foobar', 435]),
    (b'Putscript "foo" {31+}\r\n#comment\r\nInvalidSieveCommand\r\n\r\n',
     [Atom('Putscript'), 'foo', '#comment\r\nInvalidSieveCommand\r\n']),
    (b'Putscript "mysievescript" {97+}\r\n'
     b'require ["fileinto"];\r\n'
     b'\r\n'
     b'if envelope :contains "to" "tmartin+sent" {\r\n'
     b'  fileinto "INBOX.sent";\r\n'
     b'}\r\n',
     [Atom('Putscript'), 'mysievescript',
      'require ["fileinto"];\r\n'
      '\r\n'
      'if envelope :contains "to" "tmartin+sent" {\r\n'
      '  fileinto "INBOX.sent";\r\n'
      '}']),
    (b'Putscript "myforwards" {188+}\r\n'
     b'redirect "111@example.net";\r\n'
     b'\r\n'
     b'if size :under 10k {\r\n'
     b'    redirect "mobile@cell.example.com";\r\n'
     b'}\r\n'
     b'\r\n'
     b'if envelope :contains "to" "tmartin+lists" {\r\n'
     b'    redirect "lists@groups.example.com";\r\n'
     b'}\r\n',
     [Atom('Putscript'), 'myforwards',
      'redirect "111@example.net";\r\n'
      '\r\n'
      'if size :under 10k {\r\n'
      '    redirect "mobile@cell.example.com";\r\n'
      '}\r\n'
      '\r\n'
      'if envelope :contains "to" "tmartin+lists" {\r\n'
      '    redirect "lists@groups.example.com";\r\n'
      '}']),
     (b'Listscripts\r\n', [Atom('Listscripts')]),
     (b'listscripts\r\n', [Atom('listscripts')]),
     (b'Setactive "vacationscript"\r\n',
      [Atom('Setactive'), 'vacationscript']),
     (b'Setactive ""\r\n',
      [Atom('Setactive'), '']),
     (b'Setactive "baz"\r\n',
      [Atom('Setactive'), 'baz']),
     (b'Getscript "myscript"\r\n',
      [Atom('Getscript'), 'myscript']),
     (b'Deletescript "foo"\r\n',
      [Atom('Deletescript'), 'foo']),
     (b'Renamescript "foo" "bar"\r\n',
      [Atom('Renamescript'), 'foo', 'bar']),
     (b'Renamescript "baz" "bar"\r\n',
      [Atom('Renamescript'), 'baz', 'bar']),
     (b'CheckScript {31+}\r\n'
      b'#comment\r\n'
      b'InvalidSieveCommand\r\n\r\n',
      [Atom('CheckScript'),
       '#comment\r\n'
       'InvalidSieveCommand\r\n']),
     (b'NOOP\r\n', [Atom('NOOP')]),
     (b'NOOP "STARTTLS-SYNC-42"\r\n',
      [Atom('NOOP'), 'STARTTLS-SYNC-42']),
     (b'UNAUTHENTICATE\r\n', [Atom('UNAUTHENTICATE')]),

    # ManageSieve responses
    (b'OK\r\n', [Atom('OK')]),
    (b'Ok\r\n', [Atom('Ok')]),
    (b'oK\r\n', [Atom('oK')]),
    (b'ok\r\n', [Atom('ok')]),
    (b'NO\r\n', [Atom('NO')]),
    (b'No\r\n', [Atom('No')]),
    (b'nO\r\n', [Atom('nO')]),
    (b'no\r\n', [Atom('no')]),
    (b'BYE "Too many failed authentication attempts"\r\n',
     [Atom('BYE'), 'Too many failed authentication attempts']),
    (b'"IMPLEMENTATION" "Example1 ManageSieved v001"\r\n',
     ['IMPLEMENTATION', 'Example1 ManageSieved v001']),
    (b'"VERSION" "1.0"\r\n',
     ['VERSION', '1.0']),
    (b'"SASL" "PLAIN SCRAM-SHA-1 GSSAPI"\r\n',
     ['SASL', 'PLAIN SCRAM-SHA-1 GSSAPI']),
    (b'"SIEVE" "fileinto vacation"\r\n',
     ['SIEVE', 'fileinto vacation']),
    (b'"STARTTLS"\r\n', ['STARTTLS']),
    (b'NO (QUOTA/MAXSIZE) "Quota exceeded"\r\n',
     [Atom('NO'), [Atom('QUOTA/MAXSIZE')], 'Quota exceeded']),
    (b'NO "line 2: Syntax error"\r\n',
     [Atom('NO'), 'line 2: Syntax error']),
    (b'OK (WARNINGS) "line 8: server redirect action '
     b'limit is 2, this redirect might be ignored"\r\n',
     [Atom('OK'), [Atom('WARNINGS')],
      'line 8: server redirect action limit is 2, '
      'this redirect might be ignored']),
    (b'"summer_script"\r\n', ['summer_script']),
    (b'"vacation_script"\r\n', ['vacation_script']),
    (b'{13+}\r\nclever"script\r\n', ['clever"script']),
    (b'"main_script" ACTIVE\r\n',
     ['main_script', Atom('ACTIVE')]),
    (b'No (ACTIVE) "You may not delete an active script"\r\n',
     [Atom('No'), [Atom('ACTIVE')],
      'You may not delete an active script']),
    (b'No "bar already exists"\r\n',
     [Atom('No'), 'bar already exists']),
    (b'OK "NOOP completed"\r\n',
     [Atom('OK'), 'NOOP completed'])
)


RECEIVELINES: tuple[MyTest, ...] = (
    # Literals with and without continuation request
    (b'{3}\r\nfoo\r\n', ['foo']),
    (b'{3+}\r\nfoo\r\n', ['foo']),
    (b'{1}\r\n0\r\n', ['0']),
    (b'{3}\r\nfoo BAR\r\n', ['foo', Atom('BAR')]),
    (b'{1}\r\n\n', ['\n']),
    (b'{1+}\r\n\n', ['\n']),
    (b'{1}\r\n\r', ['\r']),
    (b'{1+}\r\n\r', ['\r']),
    (b'{2}\r\n\r\n', ['\r\n']),
    (b'{2+}\r\n\r\n', ['\r\n']),
    (b'{4}\r\n\nfoo', ['\nfoo']),
    (b'{4+}\r\n\nfoo', ['\nfoo']),
    (b'{4}\r\n\rfoo', ['\rfoo']),
    (b'{4+}\r\n\rfoo', ['\rfoo']),
    (b'{6}\r\n\r\nfoo', ['\r\nfoo']),
    (b'{6+}\r\n\r\nfoo', ['\r\nfoo']),
    (b'{4}\r\nfoo\n', ['foo\n']),
    (b'{4+}\r\nfoo\n', ['foo\n']),
    (b'{4}\r\nfoo\r', ['foo\r']),
    (b'{4+}\r\nfoo\r', ['foo\r']),
    (b'{6}\r\nfoo\r\n', ['foo\r\n']),
    (b'{6+}\r\nfoo\r\n', ['foo\r\n']),

    # Ridiculously complex ACAP line
    (b'(FOO NIL (NIL 0 "foo" {3}\r\nbar BAZ) 1 "foo") {3}\r\neof\r\n',
     [[Atom('FOO'), None,
       [None, 0, 'foo', 'bar', Atom('BAZ')],
       1, 'foo'], 'eof']),

    # ManageSieve responses
    (b'OK (TAG {16}\r\nSTARTTLS-SYNC-42) "Done"\r\n',
     [Atom('OK'), [Atom('TAG'), 'STARTTLS-SYNC-42'], 'Done'])
)

RECEIVELINE_ERRORS: tuple[tuple[bytes, type[Exception]], ...] = (
    # Garbage
    (b'{', ValueError),
    (b'{x}', ValueError),
    (b'}', ValueError),
    (b'(', ValueError),
    (b')', ValueError),
    (b'(()', ValueError),
    (b'0foo', ValueError),
    (b'!foo', ValueError)
)


SENDLINES: tuple[MyTest, ...] = (
    # Literals
    (b'{3+}\r\nfoo\r\n', [io.BytesIO(b'foo')]),
    (b'{3+}\r\nfoo\r\n', [io.StringIO('foo')]),
    (b'{1+}\r\n\n\r\n', [b'\n']),

    # Ridiculously complex ACAP line
    (b'(FOO NIL (NIL 0 "foo" BAZ) 1 "foo") "foo"\r\n', [
        [
            Atom('FOO'),
            None,
            [None, 0, 'foo', Atom('BAZ')],
            1,
            'foo'
        ],
        'foo'
    ]),

    # ManageSieve responses
    (b'OK (TAG "STARTTLS-SYNC-42") "Done"\r\n',
     [Atom('OK'), [Atom('TAG'), 'STARTTLS-SYNC-42'], 'Done'])

)


SENDLINE_ERRORS: tuple[tuple[Word, type[Exception]], ...] = (
    (-1, ValueError),
    (4_294_967_296, ValueError)
)


TOKENS: list[Word] = [
    None,
    Atom('atom'),
    0,
    4_294_967_295,
    '',
    ' ',
    '\0',
    '\n',
    '\r\n',
    'string',
    [],
    [[]]
]
"""Tokens to convert back-and-forth."""

# Create nested lists
TOKENS.append(copy.copy(TOKENS))


#
# Tests
#

class TestACAPConn(unittest.TestCase):
    def setUp(self):
        self.acap = MockACAP()

    def test_receiveline(self):
        acap = self.acap
        for i, (bytes_, objs) in enumerate(BYTESTOLINE + RECEIVELINES):
            with self.subTest(i=i, bytes_=bytes_, objs=objs):
                acap.file = io.BytesIO(bytes_)
                parsed = acap.receiveline()
                self.assertEqual(parsed, objs)
                for obj, cls in zip(parsed, map(type, objs)):
                    self.assertIsInstance(obj, cls)

    def test_receiveline_errors(self):
        acap = self.acap
        for i, (bytes_, exctype) in enumerate(RECEIVELINE_ERRORS):
            with self.subTest(i=i, bytes_=bytes_, exctype=exctype):
                acap.file = io.BytesIO(bytes_)
                try:
                    acap.receiveline()
                # pylint: disable=broad-exception-caught
                except Exception as exc:
                    self.assertIsInstance(exc, exctype)
                    continue
                self.fail(f'expected {exctype.__name__}')

    def test_sendline(self):
        acap = self.acap
        for i, (bytes_, objs) in enumerate(BYTESTOLINE + SENDLINES):
            with self.subTest(i=i, bytes_=bytes_, objs=objs):
                acap.file = io.BytesIO()
                rewind(objs)
                acap.sendline(*objs)    # type: ignore
                serialised = acap.file.getvalue()
                self.assertEqual(serialised, bytes_)

    def test_sendline_errors(self):
        acap = self.acap
        for i, (obj, exctype) in enumerate(SENDLINE_ERRORS):
            with self.subTest(i=i, obj=obj, exctype=exctype):
                try:
                    acap.sendline(obj)
                # pylint: disable=broad-exception-caught
                except Exception as exc:
                    self.assertIsInstance(exc, exctype)
                    continue
                self.fail(f'expected {exctype.__name__}')

    def test_sendline_partial(self):
        acap = self.acap
        for i, (bytes_, objs) in enumerate(BYTESTOLINE + SENDLINES):
            with self.subTest(i=i, bytes_=bytes_, objs=objs):
                acap.file = io.BytesIO()
                rewind(objs)
                acap.sendline(*objs, whole=False)   # type: ignore
                serialised = acap.file.getvalue()
                self.assertEqual(serialised, bytes_.removesuffix(b'\r\n'))

    @unittest.skipIf(os.getenv('QUICKTEST'), 'slow')
    def test_mutual(self):
        acap = self.acap
        for i, objs in enumerate(powerset(TOKENS)):
            with self.subTest(i=i, objs=objs):
                acap.file = io.BytesIO()
                acap.sendline(*objs)
                acap.file.seek(0)
                parsed = acap.receiveline()
                self.assertEqual(parsed, objs)


#
# Classes
#

class MockACAP(BaseACAPConn):
    file: IO[bytes] = io.BytesIO()  # type: ignore


#
# Functions
#

def powerset(elems: Sequence[T]) -> list[list[T]]:
    n = len(elems)
    return [[elems[k] for k in range(n) if i & 1 << k] for i in range(2 ** n)]


def rewind(objs):
    for obj in objs:
        if isinstance(obj, io.IOBase) and obj.seekable():
            obj.seek(0)


#
# Boilerplate
#

if __name__ == '__main__':
    unittest.main()
